#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['polyA']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['cm_to_stockholm = polyA.converters:cm_to_stockholm',
                     'polyA = polyA.__main__:main']}

setup(name='polyA',
      version='1.0.0',
      description='PolyA is a sequence annotation adjudicator.',
      author='Kaitlin Carey',
      author_email='kaitlin1.carey@umontana.edu',
      url='https://github.com/TravisWheelerLab/polyA',
      packages=packages,
      package_data=package_data,
      entry_points=entry_points,
     )
